@@ -21,10 +21,10 @@
  * along with Ocaml-mad; if not, write to the Free Software
  * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
  *
- * @author Samuel Mimram
+ * @author Samuel Mimram, David Baelde
  */
 
-/* $Id: mad_stubs.c 2395 2006-04-10 09:11:58Z smimram $ */
+/* $Id: mad_stubs.c 2694 2006-10-16 23:17:25Z dbaelde $ */
 
 #include <caml/alloc.h>
 #include <caml/callback.h>
@@ -251,7 +251,14 @@ CAMLprim value ocaml_mad_decode_frame(value madf)
     {
       data = caml_callback(mf->read_func, Val_int(read_size));
       read_size = Int_val(Field(data, 1));
-      memcpy(read_start, String_val(Field(data, 0)), read_size);
+      if (read_size>0)
+        memcpy(read_start, String_val(Field(data, 0)), read_size);
+      else
+        if (read_size=0)
+          caml_raise_constant(*caml_named_value("mad_exn_end_of_stream"));
+        else
+          caml_raise_with_arg(*caml_named_value("mad_exn_read_error"),
+              caml_copy_string("reading callback error"));
     }
     else
     {
@@ -337,8 +344,11 @@ CAMLprim value ocaml_mad_decode_frame(value madf)
    */
   dec = mad_frame_decode(&mf->frame, &mf->stream);
   if (dec)
-    if (!MAD_RECOVERABLE(mf->stream.error) && mf->stream.error != MAD_ERROR_BUFLEN)
-    {
+    if (MAD_RECOVERABLE(mf->stream.error)
+        || mf->stream.error == MAD_ERROR_BUFLEN) {
+      caml_leave_blocking_section();
+      CAMLreturn(ocaml_mad_decode_frame(madf));
+    } else {
       caml_leave_blocking_section();
       caml_raise_with_arg(*caml_named_value("mad_exn_mad_error"), caml_copy_string(mad_stream_errorstr(&mf->stream)));
     }
